""" Unit test of yaccounts package"""

import pytest
import parametrize_from_file
import yaml

from yaccounts import CsvDataAnalyzer


@pytest.fixture(scope="class")
def aa():
    aa = CsvDataAnalyzer("data.csv")
    aa.run()
    yield aa


class TestBalances:
    @parametrize_from_file.parametrize("balances.yml")
    def test_balances(self, operating_unit, year, balance, aa):
        matches = aa.find(operating_unit, year)

        assert len(matches) == 1, f"Cannot find match for {operating_unit} {year}"
        assert round(matches[0].balance) == round(
            balance
        ), f"Balance for {operating_unit} - {year} does not match"



import calendar
from collections import defaultdict
import enum
import re
import numpy as np

import pandas as pd

from .config import *


class OperatingUnitType(enum.Enum):
    PROJECT = 0
    FUND_11 = 1
    ACTUALS = 2


class OperatingUnit:
    def __init__(
        self,
        worksheet,
        operating_unit,
        year,
        name=None,
        fund_type=None,
        student_aid=None,
        hide_student_wages=False,
        hide_details=True,
    ) -> None:
        self.worksheet = worksheet
        self.operating_unit = str(operating_unit)
        self.name = name
        self.year = year
        self.student_aid = student_aid
        self.hide_student_wages = hide_student_wages
        self.hide_details = hide_details

        if fund_type is None:
            if self.operating_unit.startswith("R"):
                self.type = OperatingUnitType.PROJECT
            elif self.operating_unit.startswith("11"):
                self.type = OperatingUnitType.FUND_11
            else:
                self.type = OperatingUnitType.ACTUALS

        self.df_gen = None

        self.row_colors = {}
        self.hidden_rows = []
        self.no_sum_rows = []

        self.df = None

        if self.type == OperatingUnitType.PROJECT:
            self.previous_expenditures = self.analyzer.get_previous_expenditures(
                self.operating_unit, self.year - 1
            )

    def run(self, df):
        self.df = df
        if self.year:
            # Filter df on 'Fiscal Year' column
            self.df = self.df[self.df["Fiscal Year"] == self.year]

        if self.type in (OperatingUnitType.PROJECT, OperatingUnitType.FUND_11):
            self.verify_no_income()
        else:
            self.filter_out_budgeted_data()

        print(
            f"Running Operating Unit:{self.operating_unit} {('(' + str(self.year) + ')') if self.year else ''}",
        )

        # Add student wage data
        self.add_row("Student Wages", CODES_STUDENT_WAGES, color=COLOR_STUDENT_WAGES)
        self.create_student_wages()
        self.add_row("Student Tuition", CODES_STUDENT_TUITION, color=COLOR_STUDENT_TUITION)
        self.add_row("Faculty Spr/Sum", CODES_FACULTY_SPRING_SUMMER, color=COLOR_FACULTY)
        self.add_row("Travel", CODES_TRAVEL, color=COLOR_TRAVEL)
        self.add_detailed_rows(
            CODES_TRAVEL, ("Account Descr", "PINV Business Purpose", "JRNL Line Descr", "JH Descr")
        )

        self.add_row("Supplies/Misc", CODES_SUPPLIES, color=COLOR_SUPPLIES)
        self.add_detailed_rows(
            CODES_SUPPLIES,
            (
                "Account Descr",
                "Class Descr",
                "PINV Business Purpose",
                "JRNL Line Descr",
                "PINV Supplier Name",
            ),
        )

        self.add_row("Capital Equipment", CODES_CAPITAL, color=COLOR_CAPITAL)
        self.add_detailed_rows(
            CODES_CAPITAL,
            (
                "Account Descr",
                "Class Descr",
                "PINV Business Purpose",
                "JRNL Line Descr",
                "PINV Supplier Name",
            ),
        )

        self.add_row("Overhead", CODES_OVERHEAD, color=COLOR_OVERHEAD)
        self.row_total_expenses = self.add_row(
            "Total Expeneses", all_expense_codes(), color=COLOR_TOTAL_EXPENSES
        )

        self.add_empty_row()

        if self.type == OperatingUnitType.ACTUALS:
            self.add_row("Interest", CODES_INTEREST, actuals_coeff=-1)

        income_row_name = "Transfer Income / Carryover"
        if self.type == OperatingUnitType.PROJECT:
            income_row_name = "New Budgets or Carryover"
        elif self.type == OperatingUnitType.FUND_11:
            income_row_name = "New Budgets"

        self.add_row(
            income_row_name,
            CODES_INCOME,
            budgets_coeff=1,
            actuals_coeff=-1,
        )

        self.row_total_income = self.add_row(
            "Total Income",
            CODES_INTEREST + CODES_INCOME,
            color=COLOR_TOTAL_INCOME,
            budgets_coeff=1,
            actuals_coeff=-1,
        )
        self.add_empty_row()

        # Add a sum column that sums the values in each row, starting with the 2nd column,
        # skipping rows where "Type" is empty

        # Calculate the sum for rows where 'Type' column is not empty
        columns_to_sum = self.df_gen.columns[1:]  # Get all columns starting from the second column
        self.df_gen["Sum"] = self.df_gen.loc[
            ~self.df_gen.index.isin(self.no_sum_rows), columns_to_sum
        ].sum(axis=1)

        # Save total expenses, which is found in the self.row_total_expenses, "Sum" column
        self.total_expenses = self.df_gen.iloc[self.row_total_expenses]["Sum"]

        self.add_balance_row()
        self.add_empty_row()

        title_name = self.operating_unit
        if self.name:
            title_name = f"{self.name} ({title_name})"
        if self.year:
            title_name = f"{title_name} - {self.year}"
        self.worksheet.write_to_sheet(self.df_gen, title_name, self.hidden_rows, self.row_colors)

    def add_empty_row(self):
        # Add an empty row of None values
        empty_data = {
            "Type": [""],
            **{calendar.month_abbr[month]: [np.nan] for month in range(1, 13)},
        }

        self.df_gen = pd.concat((self.df_gen, pd.DataFrame(empty_data)), ignore_index=True)
        self.no_sum_rows.append(len(self.df_gen) - 1)

    def add_row(
        self,
        row_name,
        account_codes,
        budgets_coeff=0,
        actuals_coeff=1,
        color=None,
    ):
        """Sum the values in the 'Amount' column for each month in the given account codes.
        If actual=True, then only actual values will be summed.
        Otherwise budgeted (account income) values will be summed."""

        new_row_data = {
            "Type": [row_name],
            **{
                calendar.month_abbr[month]: [
                    budgets_coeff
                    * (
                        self.df.loc[
                            (self.df["Accounting Period"] == month)
                            & (self.df["JRNL Line Ledger"] == "BUDGETS")
                        ][COL_NAME_AMOUNT].sum()
                    )
                    + actuals_coeff
                    * (
                        self.df.loc[
                            (self.df["Accounting Period"] == month)
                            & self.df["Account"].isin(account_codes)
                            & (self.df["JRNL Line Ledger"] == "ACTUALS")
                        ][COL_NAME_AMOUNT].sum()
                    )
                ]
                for month in range(1, 13)
            },
        }

        # Add this dictionary to the dataframe
        df_new_row = pd.DataFrame(new_row_data)
        if self.df_gen is None:
            self.df_gen = df_new_row
        else:
            self.df_gen = pd.concat((self.df_gen, df_new_row), ignore_index=True)

        if color:
            self.row_colors[len(self.df_gen)] = color

        return self.df_gen.tail(1).index[0]

    def add_balance_row(self):
        if self.type == OperatingUnitType.PROJECT:
            prev_expenditures_row = len(self.df_gen)
            self.df_gen.loc[prev_expenditures_row] = {
                "Type": "Previous Expenditures",
                calendar.month_abbr[1]: self.previous_expenditures,
            }

        # Create student aid rows
        if self.student_aid:
            student_aid_row = len(self.df_gen)
            self.df_gen.loc[student_aid_row, "Type"] = "Student aid"

            non_student_aid_row = len(self.df_gen)
            self.df_gen.loc[non_student_aid_row, "Type"] = "Non-student aid"

        # Create balance row
        balance_row = len(self.df_gen)
        self.df_gen.loc[balance_row] = {"Type": "Balance (end of month)"}

        # # Iterate through columns and calcuate the balance, which is the total income minus the total expenses
        for month in range(1, 13):
            income = self.df_gen.iloc[self.row_total_income, month]
            expenses = self.df_gen.iloc[self.row_total_expenses, month]

            if month > 1:
                prev_balance = self.df_gen.iloc[balance_row, month - 1]
            elif self.type == OperatingUnitType.PROJECT:
                prev_balance = -self.df_gen.iloc[prev_expenditures_row, month]
            else:
                prev_balance = 0

            balance = income - expenses + prev_balance
            self.df_gen.iloc[len(self.df_gen) - 1, month] = balance
            if month == 12:
                self.balance = balance

            # Fill in student aid balances
            if self.student_aid:
                student_expenses = self.analyzer.get_previous_expenditures_student_aid(
                    self.operating_unit, self.year, month
                )
                student_aid_balance = self.student_aid - student_expenses
                self.df_gen.iloc[student_aid_row, month] = student_aid_balance
                self.df_gen.iloc[non_student_aid_row, month] = balance - student_aid_balance

        self.row_colors[len(self.df_gen)] = "yellow"

    def create_student_wages(self):
        # Filter to only student wages ('Account' column in STUDENT_WAGES)
        df_students = self.df[
            (self.df["Account"].isin(CODES_STUDENT_WAGES))
            & (self.df["JRNL Line Ledger"] == "ACTUALS")
        ]

        # Get unique list of value in 'JRNL Line Descr' column
        students_raw = df_students["JRNL Line Descr"].unique()
        self.students = defaultdict(list)
        for student in students_raw:
            names = re.split(", |,| ", student)
            # if not match:
            #     raise Exception(f"Could not parse student name: {student}")
            if len(names) > 1:
                student_key = f"{names[1].title()} {names[0].title()}"
            else:
                student_key = names[0]
            self.students[student_key].append(student)

        for student in self.students.copy():
            non_person = any(word in student.lower() for word in ("benefit", "overspent"))

            if non_person:
                self.students["<" + student + ">"] = self.students.pop(student)

        sorted_students = sorted(self.students.keys(), key=lambda x: (x.startswith("<"), x))
        # Create a new dictionary with the student wages by month
        student_data = {
            "Type": [f"  {s}" for s in sorted_students],
            **{
                calendar.month_abbr[month]: [
                    df_students.loc[
                        (df_students["Accounting Period"] == month)
                        & (df_students["JRNL Line Descr"].isin(self.students[student]))
                    ]["JRNL Monetary Amount -no scrn aggregation"].sum()
                    for student in sorted_students
                ]
                for month in range(1, 13)
            },
        }
        self.df_gen = pd.concat((self.df_gen, pd.DataFrame(student_data)), ignore_index=True)

        # Add new row index to hidden rows
        if self.hide_student_wages:
            self.hidden_rows.extend(
                range(len(self.df_gen) - len(student_data["Type"]) + 1, len(self.df_gen) + 1)
            )

    def add_detailed_rows(self, codes, detail_cols):
        # Filter to only travel expenses ('Account' column in TRAVEL)
        df_travel = (
            self.df[(self.df["Account"].isin(codes)) & (self.df["JRNL Line Ledger"] == "ACTUALS")]
            .rename_axis("index")
            .sort_values(by=["Accounting Period", "index"])
        )

        for index, row in df_travel.iterrows():
            data = {
                "Type": [
                    f"  Line {index+2}: {(' - '.join(str('' if pd.isnull(row[col_name]) else row[col_name] ) for col_name in detail_cols))}"
                ],
                **{
                    calendar.month_abbr[month]: (
                        row[COL_NAME_AMOUNT] if month == row["Accounting Period"] else 0
                    )
                    for month in range(1, 13)
                },
            }
            self.df_gen = pd.concat((self.df_gen, pd.DataFrame(data)), ignore_index=True)
            # self.no_sum_rows.append(len(self.df_gen) - 1)
            if self.hide_details:
                self.hidden_rows.append(len(self.df_gen))

    def verify_no_income(self):
        # Ensure df contains no CODES_INCOME
        if self.df[self.df["Account"].isin(CODES_INCOME + CODES_INTEREST)].shape[0] > 0:
            print(self.df[self.df["Account"].isin(CODES_INCOME + CODES_INTEREST)])
            raise Exception(
                f"Budgeted account {self.operating_unit} contains income. "
                "Budgeted accounts should only contain expenses."
            )

    def filter_out_budgeted_data(self):
        # Filter out budgeted data.  Only include ACTUALS or BUDGET with code in CODES_NON_BUDGETED_BUDGET_CODES
        self.df = self.df[
            (
                (self.df["JRNL Line Ledger"] == "ACTUALS")
                | (
                    (self.df["JRNL Line Ledger"] == "BUDGETS")
                    & (self.df["Account"].isin(CODES_NON_BUDGETED_BUDGET_CODES))
                )
            )
        ]

    @property
    def analyzer(self):
        return self.worksheet.workbook.analyzer



import datetime
import matplotlib
from .utils import ensure_tuple, filter_df_on_operating_units
from .operating_unit import OperatingUnit


class Worksheet:
    def __init__(self, workbook, sheet) -> None:
        self.workbook = workbook
        self.operating_units = []
        self.sheet = sheet
        self.next_start_row = 0

    def add_operating_unit(self, operating_unit, name=None, year=None, **options):
        if year is None:
            # Get all years for this operating unit from the dataframe
            years = sorted(
                self.workbook.analyzer.df[
                    self.workbook.analyzer.df["Operating Unit"] == str(operating_unit)
                ]["Fiscal Year"].unique()
            )
        else:
            years = ensure_tuple(year)

        for year in years:
            self.operating_units.append(
                OperatingUnit(self, operating_unit, year, name=name, **options)
            )

    def format_cell(self, row, col, format):
        self.sheet.conditional_format(row, col, row, col, {"type": "no_errors", "format": format})

    def format_row(self, row, format):
        self.sheet.conditional_format(row, 0, row, 13, {"type": "no_errors", "format": format})
        #     sheet.write_blank(row, i, sheet[row, i].value, color)

    def get_operating_units(self):
        return [ou.operating_unit for ou in self.operating_units]

    def run(self, df):
        for operating_uint in self.operating_units:
            operating_uint.run(filter_df_on_operating_units(df, (operating_uint.operating_unit,)))

        # Set the width of column 0 to 20
        self.sheet.set_column(0, 13, 20)

        self.sheet.set_column(1, 12, 11, cell_format=self.workbook.format_numbers)

        self.sheet.set_column(13, 13, 12, cell_format=self.workbook.format_numbers)

    def write_to_sheet(self, df, title_name, hidden_rows=(), row_colors={}):
        df.to_excel(
            self.workbook.writer,
            startrow=self.next_start_row + 1,
            index=False,
            sheet_name=self.sheet.name,
        )

        self.sheet.merge_range(
            self.next_start_row,
            0,
            self.next_start_row,
            13,
            title_name,
            cell_format=self.workbook.format_titles,
        ),

        # Hide rows
        for row in hidden_rows:
            self.sheet.set_row(self.next_start_row + row + 1, options={"hidden": True})

        # Color rows
        for row in row_colors:
            fmt = self.workbook.workbook.add_format(
                {"bg_color": matplotlib.colors.cnames[row_colors[row]]}
            )
            self.format_row(self.next_start_row + row + 1, fmt)

        self.next_start_row += len(df.index) + 2



CODES_STUDENT_WAGES = (
    5510,  # Contracts: Student teaching assist salary
    5570,  # Contracts: Student Research Assist Salary
    5572,  # Contracts: Stdnt Research Assist Sal-FICA
    5599,  # Contracts: Student Asst Salary (Reclass)
    5600,  # Time Cards: Student Wages
    5602,  # Time Cards: Student Wages-FICA
    5610,  # Time Cards: Student Wages-Teaching Assist
    5612,  # Stdnt Wages Teach Asst-FICA
    5670,  # Time Cards: Student Wages-Research Assist
    5672,  # Time Cards: Stdnt Wages Rsrch Assist-FICA
    5699,  # Time Cards: Student Wages (Reclass)
    5720,  # Non-Student Wages: Category 1 no Benefit
    5750,  # Non-Student Wages: Category 2 Part Benefit
    5812,  # Payroll Incentives (Cell Phone Personal Use)
    5950,  # Student Contract Benefits
    5960,  # Student Time Card Benefits
    5970,  # Non Student Benefits
    5999,  # Benefits (reclass)
    8905,  # Sponsored Research: Research Participant Stipends
)

CODES_STUDENT_TUITION = (
    6300,  # Scholarships
    6303,  # Graduate Tuition
    6304,  # Graduate Cash Award
    6309,  # Scholarships (reclass)
    6319,  # Student Aid (reclass)
    6390,  # Other Student Aid
)

CODES_FACULTY_SPRING_SUMMER = (
    5920,  # Faculty Supplemental Benefits
    5260,  # Faculty Sal-Spr/Sum-Benefits
    5220,  # Faculty Sal-Spr/Sum No Ben
    5240,  # Faculty Sal-Supp (Overload)
)

CODES_TRAVEL_FACULTY = (
    6190,  # Employee Development/Training
    6400,  # Insurance Expense
    6494,  # Other Medical
    7000,  # Domestic-Business Travel
    7010,  # Domestic-Recruiting Travel
    7020,  # Domestic-Prof Developmt Travel
    7050,  # Foreign-Business Travel
    7060,  # Foreign-Recruiting Travel
    7070,  # Foreign-Prof Developmt Travel
)

CODES_TRAVEL_STUDENT = (
    7030,  # Domestic-Student Travel
    7080,  # Foreign-Student Travel
)

CODES_TRAVEL = CODES_TRAVEL_FACULTY + CODES_TRAVEL_STUDENT

CODES_SUPPLIES = (
    2200,  # Accrued Liabilities - Faculty computers?
    5980,  # Other Payroll Benefits
    6000,  # Software Acquisitions/Support
    6005,  # Software / Media Licenses
    6100,  # Supplies
    6109,  # Supplies (reclass)
    6110,  # Supplies-BYU Store
    6120,  # Printing and Copying
    6125,  # Research Publications Costs
    6130,  # Postage and Mailing
    6140,  # Telecommunications
    6160,  # Advertising and Promotion
    6180,  # Hosting, Food, Entertainment
    6185,  # Guest Tickets
    6192,  # Employee Memberships, Dues
    6200,  # Contract Service-Off Campus
    6210,  # Contract Services-Campus
    6220,  # Rental Expense - Intra Campus
    6250,  # Non-capital equipment
    6255,  # Non-Cap Access Equipment
    6270,  # Equipment Maintenance
    6403,  # Other Academic Payments
    6405,  # Prizes
    6600,  # Access Equipment
    6610,  # Access Lab Equipment
    6480,  # Legal Fees
    6490,  # Other Expense
    6499,  # Other Expense (reclass)
    8940,  # Conference Hosting (R-project)
)

CODES_CAPITAL = (
    1625,  # Capital Equipment
    1725,  # AccDepre-Capital Equipment
    8930,  # Research Capital Equipment
    9126,  # AM Retirement-Assets
    9127,  # AM Retirement-Accum Depr
    9136,  # AM Transfer-Assets
    9137,  # AM Tansfer-Accum Depr
)

CODES_OVERHEAD = (
    8990,  # Facilities & Admin Recovery
    8991,  # Waived Facilities & Admin Cost
    8993,  # Subcontract Facilities & Admin
)

CODES_INTEREST = (
    4711,  # Interest Alloc-Claim on Cash
    4719,  # Non-Invest-Intrst Inc(reclass)
)

# Income only applies to non-budgeted accounts
CODES_INCOME = (
    3500,  # Beginning Net Assets
    4755,  # Distributed Royalty Income
    4891,  # Endowment Payout (UFS only)
    8920,  # Subawards w/ F&A
    8925,  # Subawards w/out F&A
    9250,  # Transfer Into Net Assets
    9260,  # Transfer Out of Net Assets
)

# Non-budgeted accounts, don't normally use BUDGET lines and they
# can be removed, except for these codes
CODES_NON_BUDGETED_BUDGET_CODES = (10,)

# These codes aren't used by anything and can be removed from the data
CODES_IGNORED = (
    1010,  # Claim on Cash
    1235,  # Sponsored Programs Receivable
    2000,  # Accounts Payable
    2210,  # Accrued Payroll-Biweekly
    3500,  # Beginning Net Assets
    4200,  # Gift Revenue-Cash
    4220,  # Grants & Contract Revenue
)

COL_NAME_AMOUNT = "JRNL Monetary Amount -no scrn aggregation"

COLOR_STUDENT_WAGES = "rosybrown"
COLOR_STUDENT_BENEFITS = "moccasin"
COLOR_STUDENT_TUITION = "moccasin"
COLOR_FACULTY = "lightskyblue"
COLOR_TRAVEL = "plum"
COLOR_SUPPLIES = "khaki"
COLOR_CAPITAL = "khaki"
COLOR_OVERHEAD = "silver"
COLOR_TOTAL_EXPENSES = "tomato"
COLOR_TOTAL_INCOME = "limegreen"


def all_expense_codes():
    return (
        CODES_STUDENT_WAGES
        + CODES_STUDENT_TUITION
        + CODES_FACULTY_SPRING_SUMMER
        + CODES_TRAVEL
        + CODES_SUPPLIES
        + CODES_OVERHEAD
        + CODES_CAPITAL
    )


def student_aid_expense_codes():
    return CODES_STUDENT_WAGES + CODES_STUDENT_TUITION + CODES_TRAVEL_STUDENT


def all_handled_codes():
    return all_expense_codes() + CODES_INTEREST + CODES_IGNORED + CODES_INCOME



import pathlib
import pandas as pd
import yaml

from .config import (
    CODES_IGNORED,
    COL_NAME_AMOUNT,
    all_expense_codes,
    all_handled_codes,
    student_aid_expense_codes,
)
from .utils import filter_df_on_operating_units
from .workbook import Workbook


class CsvDataAnalyzer:
    def __init__(self, csv_path, split_notebooks=False, yml_config=None) -> None:
        self.csv_path = csv_path

        assert isinstance(split_notebooks, bool)
        self.split_notebooks = split_notebooks

        self.yml_config = pathlib.Path(yml_config) if yml_config else None
        if self.yml_config:
            assert self.yml_config.is_file()

        self.df = pd.read_csv(self.csv_path)

        # Convert Operating Unit to string
        self.df["Operating Unit"] = self.df["Operating Unit"].apply(str)

        # Filter out ignored codes (account in CODES_IGNORED)
        self.df = self.df[~self.df["Account"].isin(CODES_IGNORED)]

        self.operating_units = sorted([str(ou) for ou in self.df["Operating Unit"].unique()])
        print(f"Operating Units Found in CSV: {', '.join(self.operating_units)}")

        self.workbooks_initialized = False
        self.workbooks = []

        self.check_unhandled_codes()

        if self.yml_config is not None:
            self._setup_workbooks_from_config()

    def add_workbook(self, path):
        self.workbooks_initialized = True
        workbook = Workbook(path)
        workbook.analyzer = self
        self.workbooks.append(workbook)
        return workbook

    def _setup_default_workbooks(self, options=None):
        if not options:
            options = self._get_default_options()

        if self.split_notebooks:
            # Create a new workbook for each operating unit
            for operating_unit in self.operating_units:
                CsvDataAnalyzer._add_operating_unit_with_options(
                    self.add_workbook(
                        pathlib.Path.cwd() / (operating_unit + ".xlsx")
                    ).add_worksheet(operating_unit),
                    operating_unit,
                    operating_unit,
                    options,
                )
        else:
            # Create a single workbook with all operating units
            wb = self.add_workbook(pathlib.Path.cwd() / "all.xlsx")
            for operating_unit in self.operating_units:
                CsvDataAnalyzer._add_operating_unit_with_options(
                    wb.add_worksheet(operating_unit), operating_unit, operating_unit, options
                )
        self.workbooks_initialized = True

    @classmethod
    def _add_operating_unit_with_options(cls, worksheet, operating_unit, name, options):
        worksheet.add_operating_unit(
            operating_unit,
            name,
            year=options["year"] if "year" in options else None,
            **{k: v for k, v in options.items() if k not in ("year",)},
        )

    def _get_default_options(self):
        return {
            "year": None,
            "student_aid": None,
            # "budgeted_account": None,
            "hide_student_wages": False,
            "hide_details": True,
        }

    def _setup_workbooks_from_config(self):
        options = self._get_default_options()

        with open(self.yml_config, "r") as fp:
            config = yaml.safe_load(fp)

        for item in [item for item in config if item not in ("workbooks",)]:
            options[item] = config[item]

        if "workbooks" not in config:
            self._setup_default_workbooks(options=options)
            return

        for workbook in config["workbooks"]:
            wb = self.add_workbook(workbook["path"])

            # Get workbook config options
            options_wb = options.copy()
            for item in [item for item in workbook if item not in ("sheets", "path")]:
                assert item in options_wb
                options_wb[item] = workbook[item]

            for worksheet in workbook["sheets"]:
                sheet = wb.add_worksheet(worksheet["name"] if "name" in worksheet else None)

                # Get worksheet config options
                options_sheet = options_wb.copy()
                for item in [item for item in worksheet if item not in ("operating_units", "name")]:
                    assert item in options_sheet
                    options_sheet[item] = worksheet[item]

                for ou in worksheet["operating_units"]:
                    # Get worksheet config options
                    options_ou = options_sheet.copy()
                    for item in [item for item in ou if item not in ("code", "name")]:
                        assert item in options_ou
                        options_ou[item] = ou[item]

                    CsvDataAnalyzer._add_operating_unit_with_options(
                        sheet, ou["code"], ou["name"] if "name" in ou else None, options_ou
                    )
        self.workbooks_initialized = True

    def run(self):
        if not self.workbooks_initialized:
            self._setup_default_workbooks()

        for workbook in self.workbooks:
            workbook.run(filter_df_on_operating_units(self.df, workbook.get_operating_units()))

    def check_unhandled_codes(self):
        # Get all account codes that were not handled
        unhandled_codes = sorted(
            self.df[
                (~self.df["Account"].isin(all_handled_codes()))
                & (self.df["JRNL Line Ledger"] == "ACTUALS")
            ]["Account"].unique()
        )

        # TODO: Add to spreadsheet instead of raising exception

        if unhandled_codes:
            raise Exception(
                f"Unhandled account codes: {','.join([str(c) for c in unhandled_codes])}.  See https://finserve.byu.edu/account-revenue-expense"
            )

    def get_all_operating_units(self):
        return [
            ou
            for workbook in self.workbooks
            for worksheet in workbook.worksheets
            for ou in worksheet.operating_units
        ]

    def find(self, operating_unit, year=None):
        operating_uints = [
            ou for ou in self.get_all_operating_units() if ou.operating_unit == str(operating_unit)
        ]
        if year:
            operating_uints = [ou for ou in operating_uints if ou.year == year]
        return operating_uints

    def get_previous_expenditures(self, operating_unit, year):
        """Return the sum of expenditures for this year and all previous years"""
        return self.df[
            (self.df["Operating Unit"] == operating_unit)
            & (self.df["Fiscal Year"] <= year)
            & (self.df["JRNL Line Ledger"] == "ACTUALS")
            & (self.df["Account"].isin(all_expense_codes()))
        ][COL_NAME_AMOUNT].sum()

    def get_previous_expenditures_student_aid(self, operating_unit, year, month):
        """Return the sum of student aid expenditures up to and including the given month"""

        return self.df[
            (self.df["Operating Unit"] == operating_unit)
            & (self.df["JRNL Line Ledger"] == "ACTUALS")
            & (self.df["Account"].isin(student_aid_expense_codes()))
            & (
                (self.df["Fiscal Year"] < year)
                | ((self.df["Fiscal Year"] == year) & (self.df["Accounting Period"] <= month))
            )
        ][COL_NAME_AMOUNT].sum()



def filter_df_on_operating_units(df, operating_units):
    return df[df["Operating Unit"].isin(operating_units)]


def ensure_tuple(value):
    if isinstance(value, tuple):
        return value
    elif isinstance(value, list):
        return tuple(value)
    else:
        return (value,)



from setuptools import setup

setup(
    name="yaccounts",
    packages=["yaccounts"],
    version="1.2.3",
    description="Parsing scripts for BYU's transaction details",
    author="Jeff Goeders",
    author_email="jeff.goeders@gmail.com",
    license="MIT",
    # url="https://github.com/byu-cpe/ygrader",
    install_requires=["pandas", "openpyxl", "xlsxwriter", "matplotlib", "pyyaml"],
)
